  function [X, Y, F, timecost, tt, nftime, quadtime, T] = driver_MaxRQ(A, method, fns, params)
% driver for Maximum of multiple Rayleigh quotients.
% 
% INPUT:
% A: a n by N matrix. A column is a point in n dimensional space. The number of points is N.
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X        : all iterates of x
% Y        : all iterates of y
% F        : function values of all iterates
% T        : accumulated computational time for iterations
% timecost : total computational time
% tt       : number of function evaluations
% nftime   : time of gradient evaluations
% quadtime : number of retraction actions
% 
% By 

    fns.f = @(x)f(x, A);
    fns.gf = @(x,value)gf(x, A, value);
    fns.gf_exists = @(x)gf_exists(x, A);
   
   if(params.manifold == 2)
    if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
        fns.f = @(x)f_intr(x, A);
        fns.gf = @(x,value)gf_embedded_intr(x, A, value);
    end
   end
  
   

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, Y, F, timecost, tt, nftime, quadtime, T] = RQNBM(fns, params);
    elseif(method == 2)
        [X, Y, F, timecost, tt, nftime, quadtime,T] = RPBM(fns, params);
    elseif(method == 3)
        [X, Y, F, timecost, tt, nftime, quadtime,T] = RPBM(fns, params);
    elseif(method == 4)     
        [X, Y, F, timecost, tt, nftime, quadtime,T] = REsubGA(fns, params);
    elseif(method == 5)
        [X, Y, F, timecost, tt, nftime, quadtime, T] = LRQNBM(fns, params);
    end
end

function output = f(x, A)
    N = size(A, 3);  
    Temp = zeros(N, 1);  
    for i = 1 : N
        Temp(i) = 0.5 * x.U' * (A{i} * x.U);
    end
    output.f = max(Temp);
end

function output = f_intr(x, A)
    N = size(A, 3);
    Temp = zeros(N, 1); 
    for i = 1 : N
        Temp(i) = 0.5 * x.U' * (A{i} * x.U);
    end
    output.f = max(Temp);
end



function output = gf(x, A, value)
   if(~isfield(value, 'f'))
        value = f_intr(x, A);
    end
   N = size(A, 3);  
   temp = 0;
   k = 0;
   for i = 1 : N 
       value_i = 0.5 * x.U' * (A{i} * x.U);
       if(value.f == value_i)
            temp = temp + A{i} * x.U;   
            k = k + 1;
       end       
   end
   temp = temp / k;   
     %output = 2 * (eye(size(x.U, 1)) - x.U * x') * temp; %% small numerical error but slow
   output.TV = temp - x.U * (x.U' * temp); %% big numerical error, but fast
end

function output = gf_embedded_intr(x, A, value)
    N = size(A, 3); 
   if(~isfield(value, 'f'))
        value = f_intr(x, A);
   end
   temp = 0;
   k = 0;
    for i = 1 : N         
        value_i = 0.5 * x.U' * (A{i} * x.U);
        if(value.f == value_i)
            temp = temp + A{i} * x.U;
            k = k + 1;  
            break;
        end 
    end
    temp = temp / k;
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    OmegaK = HHRTeta(x.U, temp, x.HHR, x.taus);
    K = OmegaK(p+1:end, :);
    output.TV = zeros(n-1, 1);
    output.TV(1 : end) = reshape(K, [], 1);
end

function output = gf_exists(x, A)
    output = 1;
end

function x = Get_info_intr(x)
    [x.HHR, x.taus] = QRmatrices(x.U);
end

function output = embedded_intr_c(x, eta)
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    OmegaK = HHRTeta(x.U, eta.TV, x.HHR, x.taus);
    omega = OmegaK(1:p, :);
    omega = (omega - omega') / 2;
    K = OmegaK(p+1:end, :);
    r2 = sqrt(2);
    omega = r2 * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = embedded_full_c(x, v)
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    r2 = sqrt(2);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega - omega';
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = HHReta(x.U, [omega; K], x.HHR, x.taus);
end

function [HHR, taus] = QRmatrices(U)
    [n, k] = size(U);
    HHR = zeros(n, k);
    taus = zeros(k, 1);
    A = U;
    
    for i = 1 : k
        HHR(i : n, i) = A(i : n, i);
        HHR(i, i) = HHR(i, i) + norm(A(i : n, i)); %sign(A(i, i)) * 
        taus(i) = 2 / (HHR(i : n, i)' * HHR(i : n, i));
        A = A - taus(i) * HHR(:, i) * (HHR(:, i)' * A);
    end
end%%----------

function output = HHRTeta(U, eta, HHR, taus)
    [n, k] = size(U);
    output = eta;
    for i = 1 : k
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end%%-------------

function output = HHReta(U, eta, HHR, taus)
    [n, k] = size(U);
    output = eta;
    for i = k : -1 : 1
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end