function [X, Y, F, timecost, tt ,nftime, quadtime, T] = driver_OBB(A, method, fns, params)
% driver for Oriented Bounding Box problem.
% See https://www.elen.ucl.ac.be/Proceedings/esann/esannpdf/es2010-117.pdf
% 
% INPUT:
% A: a n by n matrix. A column is a point in n dimensional space. The number of points is n.
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates of x
% Y : all iterates of y
% tt : number of function evaluations
% F : function values of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nftime : number of function evaluations
% quadtime : number of retraction actions
% 
% By

    fns.f = @(x)f(x, A);
    if(params.vector_transport == 7 || params.vector_transport == 13)
        fns.gf = @(x,value)gf_intr(x, A,value);
    else
        fns.gf = @(x,value)gf(x, A,value);
    end
    fns.gf_exists = @(x)gf_exists(x, A);

%     nH = 0;
%     nV = 0;
%     nVp = 0;
   if(method == 1)
        [X, Y, F, timecost, tt, nftime, quadtime, T] = RQNBM(fns, params);
    elseif(method == 2)       
        [X, Y, F, timecost, tt] = RPBM(fns, params);
    elseif(method == 3)
        [X, Y, F, timecost, tt] = RPBM(fns, params);
    elseif(method == 4)
        [X, Y, F, timecost, tt] = REsubGA(fns, params);
    elseif(method == 5)
        [X, Y, F, timecost, tt, nftime, quadtime] = LRQNBM(fns, params);
    end
end

function [output, x] = f(x, A)
    xx = x.U * A;
    max_xx = max(xx');
    min_xx = min(xx');
    diff = max_xx - min_xx;
    temp = 1;
    for i = 1 : length(diff)
        temp = diff(i) * temp;
    end
    output.f = temp;
end

function ans = gf(x, A, value)  % gradient of f(x)
    nn = size(A, 1);
    xx = x.U * A;
    [max_xx, max_index] = max(xx');
    [min_xx, min_index] = min(xx');
    diff = max_xx - min_xx;
    output = zeros(size(A));
    total = 1;
    for i = 1 : nn
        total = total * diff(i);
    end
    
    for i = 1 : nn
        output(i, max_index(i)) = total / diff(i);
    end
    
    for i = 1 : nn
        output(i, min_index(i)) = - total / diff(i);
    end
    temp = output * A';
    ans.TV = temp - x.U * (x.U' * temp + temp' * x.U) / 2;
end

function [output, x] = gf_intr(x, A, value)  % gradient of f(x)
    nn = size(A, 1);
    xx = x.U * A;
    [max_xx, max_index] = max(xx');
    [min_xx, min_index] = min(xx');
    diff = max_xx - min_xx;
    Temp = zeros(size(A));
    total = 1;
    for i = 1 : nn
        total = total * diff(i);
    end
    
    for i = 1 : nn
        Temp(i, max_index(i)) = total / diff(i);
    end
    
    for i = 1 : nn
        Temp(i, min_index(i)) = - total / diff(i);
    end
    Q = Temp * A';
    [n, p] = size(x.U);
    omega = 0.5 * x.U' * Q;
    omega = omega - omega';
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    output.TV(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
end

function output = gf_exists(x, A)
    output = 1;
end
