function [M, Y, F, timecost,tt,nftime, quadtime] = driver_RGM(X, fns, method, params)
% driver for The geometric median problem.
% 
% INPUT:
% X     : a n by m matrix. A column is a point in n dimensional space. The number of points is m .
% method: indication of method
% fns   : a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% M        : all iterates of x
% Y        : all iterates of y
% F        : function values of all iterates
% T        : accumulated computational time for iterations
% timecost : total computational time
% tt       : number of function evaluations
% nftime   : time of gradient evaluations
% quadtime : number of retraction actions
% 

    fns.f = @(x)f(x, X);
    fns.gf = @(x, value)gf(x, X,value);
    if(params.manifold == 2)
    if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
        fns.f = @(x)f_intr(x, X);
        fns.gf = @(x,value)gf_intr(x, X,value);
    end
   end

    if(method == 1)
        [M, Y, F, timecost,tt, nftime, quadtime] = RQNBM(fns, params);
    elseif(method == 2)
        [M, Y, F, timecost,tt,nftime,quadtime] = RPBM(fns, params);        
    elseif(method == 3)
        [M, Y, F, timecost,tt,nftime,quadtime] = RPBM(fns, params);
    elseif(method == 4)
        [M, Y, F, timecost,tt,nftime,quadtime] = REsubGA(fns, params);
    elseif(method == 5)
        [M, Y, F, timecost,tt, nftime, quadtime] = LRQNBM(fns, params);
    end
end

function output = f(x, X)
    temp1 = 0;
    m = size(X,2);
    for i = 1 : m 
        temp = dist(X{i},x);
        temp1 = temp1 + temp;
    end
    output.f = temp1 / m;
end

function output = f_intr(x, X)
    temp1 = 0;
    m = size(X,2);
    for i = 1 : m 
        temp = dist(X{i},x);
        temp1 = temp1 + temp;
    end
    output.f = temp1 / m;
end

function output = gf(x, X, value)
   n=size(x.U,1);
   [p,m] = size(X);  
   output.TV = zeros(n,1);
   for i = 1 : m       
       dis = dist(X{i},x);
       log = logarithm(x,X{i});
       output.TV = output.TV - log.TV/m /dis;
   end
end

function output = gf_intr(x, X, value)
   [p,m] = size(X); 
   n=size(x.U,1);
   temp = zeros(n,1);
   for i = 1 : m
       dis = dist(X{i},x);
       log = logarithm(x,X{i});
       temp = temp - log.TV/m /dis;
       if(~isfield(x, 'HHR'))
           x = Get_info_intr(x);
       end
       [n, p] = size(x.U);
       OmegaK = HHRTeta(x.U, temp, x.HHR, x.taus);
       K = OmegaK(p+1:end, :);
       output.TV = zeros(n-1, 1);
       output.TV(1 : end) = reshape(K, [], 1);
   end
end

function output = gf_exists(x, A)
    output = 1;
end

  function v = logarithm(x1, x2)
        proj = @(x, d) d - x*(x(:)'*d(:));
        v.TV = proj(x1.U, x2.U - x1.U);
        di = dist(x1, x2);
        % If the two points are "far apart", correct the norm.
        if di > 1e-6
            nv = norm(v.TV, 'fro');
            v.TV = v.TV * (di / nv);
        end
  end
    
 
    
   function d = dist(x, y)
        
        % The following code is mathematically equivalent to the
        % computation d = acos(x(:)'*y(:)) but is much more accurate when
        % x and y are close.
        
        chordal_distance = norm(x.U - y.U, 'fro');
        d = real(2*asin(.5*chordal_distance));
        
        % Note: for x and y almost antipodal, the accuracy is good but not
        % as good as possible. One way to improve it is by using the
        % following branching:
        % % if chordal_distance > 1.9
        % %     d = pi - dist(x, -y);
        % % end
        % It is rarely necessary to compute the distance between
        % almost-antipodal points with full accuracy in Manopt, hence we
        % favor a simpler code.
        
   end
    
  
function x = Get_info_intr(x)
    [x.HHR, x.taus] = QRmatrices(x.U);
end

function output = embedded_intr_c(x, eta)
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    OmegaK = HHRTeta(x.U, eta.TV, x.HHR, x.taus);
    omega = OmegaK(1:p, :);
    omega = (omega - omega') / 2;
    K = OmegaK(p+1:end, :);
    r2 = sqrt(2);
    omega = r2 * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = embedded_full_c(x, v)
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    r2 = sqrt(2);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega - omega';
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = HHReta(x.U, [omega; K], x.HHR, x.taus);
end

function [HHR, taus] = QRmatrices(U)
    [n, k] = size(U);
    HHR = zeros(n, k);
    taus = zeros(k, 1);
    A = U;
    
    for i = 1 : k
        HHR(i : n, i) = A(i : n, i);
        HHR(i, i) = HHR(i, i) + norm(A(i : n, i)); %sign(A(i, i)) * 
        taus(i) = 2 / (HHR(i : n, i)' * HHR(i : n, i));
        A = A - taus(i) * HHR(:, i) * (HHR(:, i)' * A);
    end
end%%----------

function output = HHRTeta(U, eta, HHR, taus)
    [n, k] = size(U);
    output = eta;
    for i = 1 : k
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end%%-------------

function output = HHReta(U, eta, HHR, taus)
    [n, k] = size(U);
    output = eta;
    for i = k : -1 : 1
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end