function [X, Y, F, timecost, tt, nftime, quadtime,T] = driver_SV(Q, method, fns, params)
% driver for the sparse vector problem.
% 
% INPUT:
% Q     : a n by n matrix.
% method: indication of method
% fns   : a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X        : all iterates of x
% Y        : all iterates of y
% F        : function values of all iterates
% T        : accumulated computational time for iterations
% timecost : total computational time
% tt       : number of function evaluations
% nftime   : time of gradient evaluations
% quadtime : number of retraction actions
% 
% By Wen Huang

    fns.f = @(x)f(x, Q);
    fns.gf = @(x,value)gf(x, Q,value);
    fns.gf_exists = @(x,value)gf_exists(x, Q,value);
   
   if(params.manifold == 2)
    if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
        fns.f = @(x)f_intr(x, Q);
        fns.gf = @(x,value)gf_embedded_intr(x, Q,value);
    end
   end
  

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, Y, F, timecost, tt, nftime, quadtime,T] = RQNBM(fns, params);
    elseif(method == 2)
        [X, Y, F, timecost, tt, nftime, quadtime,T] = RPBM(fns, params); 
    elseif(method == 3)
        [X, Y, F, timecost, tt, nftime, quadtime,T]  = RPBM(fns, params);
    elseif(method == 4)
        [X, Y, F, timecost, tt, nftime, quadtime,T] = REsubGA(fns, params);
    elseif(method == 5)
        [X, Y, F, timecost, tt, nftime, quadtime] = LRQNBM(fns, params);    
    end
end

function output = f(x, Q)
    temp = Q * x.U;
    output.f = norm( temp ,1);
end 

function output = f_intr(x, Q)
     temp  = Q * x.U;
    output.f = norm( temp,1);
end


function output = gf(x, Q, value)
   temp = sign(Q * x.U);
   temp = Q' * temp;  
%     output = 2 * (eye(size(x, 1)) - x * x') * temp; %% small numerical error but slow
    output.TV = temp - x.U * (x.U' * temp); %% big numerical error, but fast
end

function output = gf_embedded_intr(x, Q,value)
%    n=size(x.U);
   temp = sign(Q * x.U);
   temp = Q' * temp;
   if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
   end
    [n, p] = size(x.U);
    OmegaK = HHRTeta(x.U, temp, x.HHR, x.taus);
    K = OmegaK(p+1:end, :);
    output.TV = zeros(n-1, 1);
    output.TV(1 : end) = reshape(K, [], 1);
end

function output = gf_exists(x, A)
    output = 1;
end

function x = Get_info_intr(x)
    [x.HHR, x.taus] = QRmatrices(x.U);
end

function output = embedded_intr_c(x, eta)
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    OmegaK = HHRTeta(x.U, eta.TV, x.HHR, x.taus);
    omega = OmegaK(1:p, :);
    omega = (omega - omega') / 2;
    K = OmegaK(p+1:end, :);
    r2 = sqrt(2);
    omega = r2 * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = embedded_full_c(x, v)
    if(~isfield(x, 'HHR'))
        x = Get_info_intr(x);
    end
    [n, p] = size(x.U);
    r2 = sqrt(2);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega - omega';
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = HHReta(x.U, [omega; K], x.HHR, x.taus);
end

function [HHR, taus] = QRmatrices(U)
    [n, k] = size(U);
    HHR = zeros(n, k);
    taus = zeros(k, 1);
    A = U;
    
    for i = 1 : k
        HHR(i : n, i) = A(i : n, i);
        HHR(i, i) = HHR(i, i) + norm(A(i : n, i)); %sign(A(i, i)) * 
        taus(i) = 2 / (HHR(i : n, i)' * HHR(i : n, i));
        A = A - taus(i) * HHR(:, i) * (HHR(:, i)' * A);
    end
end%%----------

function output = HHRTeta(U, eta, HHR, taus)
    [n, k] = size(U);
    output = eta;
    for i = 1 : k
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end%%-------------

function output = HHReta(U, eta, HHR, taus)
    [n, k] = size(U);
    output = eta;
    for i = k : -1 : 1
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end