function varargout = LRQNBM(fns, params)
%Solver of Riemannian quasi-newton bundle method
% fns
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x1 M to T_{R_x1(d)} M. Here x2 = R_x1(d).
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%                                  and return some parameters which are useful in fns.Tranv_Params.
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
%     params.H0 : initial approximation of inverse Hessian.
% not required:
%     params.err [1e-5]         : tolerance of stopping criterion.
%     params.error [1e-4]       : another tolerance of stopping criterion
%     params.m [8]              : number of s and u for storage
%     params.max_t [5000]       : the maximum number of iterations.
%     params.update [1]         : '0' means that updates to Hk are skipped, '1' means update Hk.
%     params.rho [0.1 10-3]     : correction parameter
%     params.minf               : the known optimal objective function value or the minimum value found by the used algorithms
%     params.E [1]              : control direction length
%     params.mu0 [0.18]         : injectivity radius
%     params.Gamma [1000]       : set the parameter to reduce SR1 updates
%     params.maxstepsize [1]    : allowed maximum stepsize 
%

% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates of x
%     varargout{2} : all iterates of y
%     varargout{3} : function values of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : time of function evaluations
%     varargout{7} : time to solve a quadratic programming problem
%     varargout{8} : accumulated computational time for iterations
%     varargout{9} : part of the values of wk


    fprintf('M-RQNBM\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'm', 8, 'int', 0, inf);
    params = set_default_para(params, 'error', 1e-4, 'float', 0, 1);
    params = set_default_para(params, 'err', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'rho', 0.1, 'float', 0, 1);
    params = set_default_para(params, 'max_t', 5000, 'int', 1, inf);
    params = set_default_para(params, 'update', 1, 'int', 0, 1);
    params = set_default_para(params, 'maxstepsize', 1, 'float', 1, inf);
    params = set_default_para(params, 'mu0', 0.18, 'float', 0, inf);
    params = set_default_para(params, 'E', 1, 'float', 0, inf);
    params = set_default_para(params, 'Gamma', 1000, 'int', 1, inf);

    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))  % isfield (s，'fieldname')检查结构体S是否包含由fieldname指定域
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.Hv = fs.Hv;  %
        fns.R = fs.R;
        fns.inpro = fs.inpro;
        fns.Tranv = fs.Tranv;
        fns.beta = fs.beta;
        fns.invTranv = fs.invTranv;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'Hv');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'Tranv');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');

    % Check whether required parameters are legal
    if(~isfield(params, 'x0') || ~isfield(params, 'H0'))
        error('Invalid arguments: missing initial x0 or H0');
    end
    % Initialization
    times = 0;
    tt =1;
    m = params.m;
    x1 = params.x0;
    H = params.H0;
    n = size(H,2);
    D = params.D;
    maxstepsize = params.maxstepsize;
    f1 = fns.f(x1);
    num = 0;
    gradfx1 = fns.gf(x1,f1);
    cgradfx1 = gradfx1;
    w1 = fns.inpro(x1, cgradfx1, cgradfx1); 
    W(1) = w1;
    y1 = x1;
    c_alpha = 0;
    X{1} = x1;
    Y{1} = y1;
    F(1) = f1;
    G{1} = gradfx1;
    iC = 0;
    nftime = 0;
    quadtime = 0;
    RHO = []; 
    T(1)=0;
%     Q(1)=0;
%     NF(1)=0;
%     OT(1)=0;
    qtime = 0;
    start_index = 1;
    S = {};
    U = {};
    SS = [];
    SU = [];
    UU = [];
    US = [];
    gamma1 = 1;
    t_L = -1;
    nn = 0;
    fprintf('times : %d, f(x) : %e\n', times, f1.f);
    options1 = optimset('Display','off','TolX', 1e-5, 'TolFun', 1e-5, 'Algorithm', 'active-set');
    options2 = optimset('Display','off','TolX', 1e-6, 'TolFun', 1e-6, 'Algorithm', 'active-set');
    tic 
    while(times < params.max_t)

        if((f1.f - params.minf) / (abs(params.minf) + 1) <= params.error)
            break;
        end
        % search direction 
        if(t_L == 0) 
            eta = Hv(fns, x1, S, U, UU, US, cgradfx1, gamma1, start_index);
            eta = fns.STV(x1, -1, eta);  
        elseif(t_L>0)
            eta = Get_eta(cgradfx1, x1, S, U, RHO, start_index, gamma1, m, fns);
            if(length(S) >0 && inp_u_s >= params.rho) % discard 
                S(end)=[];
                U(end)=[];
                RHO(end) = [];
            end
        else
            eta = fns.STV(x1, -1, cgradfx1);
        end
        if(times >= params.t) %  
            params.rho = 1e-3;
        end
        w1 = fns.inpro(x1, eta, fns.STV(x1, -1, cgradfx1)) + 2 * c_alpha;

        if(w1 < params.err)  % stopping criterion
           break;
        end
                                      
        n_eta = sqrt(fns.inpro(x1, eta, eta));  % norm of search direction 
        initstepsize = min([maxstepsize, params.mu0/n_eta]);  % initial stepsize
        
        % line search
        [new_eta, Tgfnew, t_R, t_L, alpha, x2, y2, fx2, fy2, gfnew, ins, beta, lnf, lnftime] = linesearch_R(fns, eta, x1, f1, w1, n_eta, initstepsize, params, D);
      
        if(t_L>0) % serious step
            gradfx1 = gfnew;
            G={};
            HG = {};
            G{1} = gradfx1;
            G1(:,1) = reshape(gradfx1.TV,[],1);
            
            if(params.update== 1)
                [Tg, new_eta, x1, y2] = fns.Tranv(x1, new_eta, y2, gradfx1); 
                u_k = fns.TVaddTV(y2, gfnew, fns.STV(y2, -1, Tg)); 
                s_k = fns.Tranv(x1, new_eta, x2, new_eta);     
                inp_u_s = fns.inpro(x2, u_k, s_k);  
                inp_u_u = fns.inpro(x2, u_k, u_k);
                gamma11 = inp_u_s / inp_u_u;
                rho1 = 1 / inp_u_s;

                if(inp_u_s >= params.rho)
                    gamma1 = gamma11;
                   % transport and store new.
                    U{length(U) + 1} = u_k;
                    S{length(S) + 1 } = s_k;
                    RHO(length(RHO) + 1) = rho1;
                    for i = 1 : length(S)
                        U{i} = fns.Tranv(x1, new_eta, x2, U{i});
                        S{i} = fns.Tranv(x1, new_eta, x2, S{i});
                    end                       
                else
                    for i = 1 : length(S)                      
                        U{i} = fns.Tranv(x1, new_eta, x2, U{i});
                        S{i} = fns.Tranv(x1, new_eta, x2, S{i});
                    end
                    iU = 0;
                end
            else
%                 for i = 1 : length(S)
%                     U{i} = fns.Tranv(x1, new_eta, x2, U{i});
%                     S{i} = fns.Tranv(x1, new_eta, x2, S{i});
%                 end
            end
            cgradfx1 = gfnew;
            c_alpha = 0;         
            nn = 1;
        
        else
            if (nn == 1)     
                eta = Hv(fns, x2, S, U, UU, US,  cgradfx1, gamma1, start_index); 
                eta = fns.STV(x2, -1, eta);
                new_eta = fns.STV(x2,t_R, eta);
            end
          
            G{2} = Tgfnew;
            cgradfx = cgradfx1;
            re_cgradfx1 = reshape(cgradfx1.TV,[],1);
            if(G{1}.TV~=re_cgradfx1)
                G{3} = cgradfx1;
                HG  = HV(fns, x2, S, U, UU, US, G, gamma1, start_index);
                for i = 1 : 3
                    HG1(:,i) = reshape(HG{i}.TV,[],1);
                end            

                for i= 1 : 3
                    G1(:,i) = reshape(G{i}.TV,[],1);
                end
                H1=  G1'* HG1;
                H1 =(H1 + H1')/2;
                c = [0; alpha; c_alpha];
                e = ones(1,3);  %
                w0 = ones(3,1)/3;
                O = zeros(3,1);             
                quadtimestart = toc;
                lambdav = quadprogsub(H1, c, [], [], e, 1, O, [], w0, options1);  
                qtime =  qtime+ toc - quadtimestart;
                quadtime =quadtime + toc - quadtimestart ;
                c_alpha = lambdav' * c;
                lambda.TV = reshape(lambdav,[],1);           
                cgradfx1 = fns.Hv(x2, G1, lambda);
                G(:,3) = [];
                HG1(:,3) = [];
                G1(:,3) = [];
                HG(3) = [];
            else
                HG = HV(fns, x2, S, U, UU, US, G, gamma1, start_index);
                for i = 1 : 2 
                    HG1(:,i) = reshape(HG{i}.TV,[],1);
                end            

                for i = 1:2
                    G1(:,i) = reshape(G{i}.TV,[],1);
                end
                H1=  G1'* HG1;
                H1 =(H1 + H1')/2;
                c = [0; alpha];
                e = ones(1,2);  %
                w0 = ones(2,1)/2;
                O = zeros(2,1);
                quadtimestart = toc;
                lambdav = quadprogsub(H1, c, [], [], e, 1, O, [], w0, options2);  %
                qtime =  qtime+ toc - quadtimestart;
                quadtime =quadtime + toc - quadtimestart ;
                c_alpha = lambdav' * c;
                lambda.TV = reshape(lambdav,[],1);           
                cgradfx1 = fns.Hv(x2, G1, lambda);

            end
            if(params.update ==1)
                u_k = fns.TVaddTV(x2, Tgfnew, fns.STV(x1, -1, gradfx1));                            
                Huktemp  = fns.TVaddTV(x2, HG{2}, fns.STV(x2,-1,HG{1}));
                v_k = fns.TVaddTV(x2, Huktemp, fns.STV(y2, -1, new_eta));
                inp_u_v = fns.inpro(x2, u_k, v_k);
                inp_cg_v = fns.inpro(x2, cgradfx, v_k);               
                inp_u_s = fns.inpro(x2, u_k, new_eta);   
                rho1 = 1 / inp_u_s;
            end
            n_cgradfx1 = fns.inpro(x2, cgradfx1, cgradfx1);
            if(params.update== 1)
                if( inp_cg_v < 0 &&( nn==1 ||(iC==0 && params.rho * n_cgradfx1 <= inp_cg_v^2 / inp_u_v && params.rho * n <= fns.inpro(x2, v_k, v_k) / inp_u_v)))
%                 if( inp_cg_v < 0 &&( params.rho * n_cgradfx1 <= inp_cg_v^2 / inp_u_v && params.rho * n <= fns.inpro(x2, v_k, v_k) / inp_u_v))
                    gamma1 = 1;

                    if(length(S) < m)
                        S{length(S) + 1} = new_eta;
                        U{length(U) + 1} = u_k;
                        RHO(length(RHO) + 1) = rho1;
                    else
                        S{start_index} = new_eta;
                        U{start_index} = u_k;
                        RHO(start_index) = rho1;
                        start_index = start_index + 1;
                        if(start_index > m)
                            start_index = 1;
                        end
                    end
                    [SS, SU, UU, US] = updateSS_SY(fns, x2, S, U, SS, SU, UU, US, start_index);
                    iU = 1;
                else
                   iU = 0;
                end  
            end
            nn = 0;
        end             
        if(times >= params.Gamma)  %
            iC = 1;
        end
     
        tt = tt + lnf ; 
        nftime = nftime + lnftime;
        times = times + 1;  
        W1(times + 1) = w1;
      
        if(x1.U ~= x2.U)    
          num =num +1;
          x1 = x2;
          y1 = y2;
          f1 = fx2;      
          X{num + 1} = x1;
          Y{num + 1} = y1;
          F(num + 1) = f1;
          T(num + 1) = toc;
          W(num + 1) = W1(times);
%         Q(num + 1) = quadtime;
%         NF(num + 1) = nftime;
          qtime = 0;
%         OT(num + 1) = toc;
        end 
    end
    timecost = toc;
    varargout{1} = X;
    varargout{2} = Y;  
    varargout{3} = F; 
    varargout{4} = timecost;
    varargout{5} = tt;
    varargout{6} = nftime;
    varargout{7} = quadtime;
    varargout{8} = T;
    varargout{9} = W;
%     varargout{10} = Q;
%     varargout{11} = NF;
%     varargout{12} = OT;
end   

function [SSnew, SYnew, YYnew, YSnew] = updateSS_SY(fns, x, S, Y, SS, SY, YY, YS, start_index)
    n1 = length(S);
    n2 = size(YY, 1);
    SSnew = zeros(n1, n1);
    SYnew = zeros(n1, n1);
    YYnew = zeros(n1, n1); 
    YSnew = zeros(n1, n1);
    if(n2 < n1)
%         SSnew(1 : n2, 1 : n2) = SS;  
%         SYnew(1 : n2, 1 : n2) = SY;
        YYnew(1 : n2, 1 : n2) = YY;
        YSnew(1 : n2, 1 : n2) = YS;
%         SSnew(n1, n1) = fns.inpro(x, S{n1}, S{n1});
%         SYnew(n1, n1) = fns.inpro(x, S{n1}, Y{n1});
        YYnew(n1, n1) = fns.inpro(x, Y{n1}, Y{n1});
        YSnew(n1, n1) = fns.inpro(x, S{n1}, Y{n1});
        for i = 1 : n2
%             SYnew(n1, i) = fns.inpro(x, S{n1}, Y{i});
%             SYnew(i, n1) = SYnew(n1, i);
%             SSnew(n1, i) = fns.inpro(x, S{n1}, S{i});
%             SSnew(i, n1) = SSnew(n1, i);
            YYnew(i, n1) = fns.inpro(x, Y{n1}, Y{i});
            YYnew(n1, i) = YYnew(i, n1);
            YSnew(i, n1) = fns.inpro(x, Y{n1}, S{i});
            YSnew(n1, i) = YSnew(i, n1);
        end
    else
        %n1 = m in this case
%         SSnew(1 : n1 - 1, 1 : n1 - 1) = SS(2 : n1, 2 : n1);
%         SYnew(1 : n1 - 1, 1 : n1 - 1) = SY(2 : n1, 2 : n1);
        YYnew(1 : n1 - 1, 1 : n1 - 1) = YY(2 : n1, 2 : n1);
        YSnew(1 : n1 - 1, 1 : n1 - 1) = YS(2 : n1, 2 : n1);
        end_ind = start_index - 1;
        if(end_ind < 1)
            end_ind = end_ind + n1;
        end
%         SSnew(n1, n1) = fns.inpro(x, S{end_ind}, S{end_ind});
%         SYnew(n1, n1) = fns.inpro(x, S{end_ind}, Y{end_ind});
        YYnew(n1, n1) = fns.inpro(x, Y{end_ind}, Y{end_ind});
        YSnew(n1, n1) = fns.inpro(x, S{end_ind}, Y{end_ind});
        for i = 1 : n1 - 1
            ind = i - 1 + start_index;
            if(ind > n1)
                ind = ind - n1;
            end
%             SYnew(n1, i) = fns.inpro(x, S{end_ind}, Y{ind});
%             SYnew(i, n1) = SYnew(n1, i);
%             SSnew(n1, i) = fns.inpro(x, S{end_ind}, S{ind});
%             SSnew(i, n1) = SSnew(n1, i);
            YYnew(i, n1) = fns.inpro(x, Y{end_ind}, Y{ind});
            YYnew(n1, i) = YYnew(i, n1);
            YSnew(i, n1) = fns.inpro(x, Y{end_ind}, S{ind});
            YSnew(n1, i) = YSnew(i, n1);
        end
    end
end


function output = Hv(fns, x, S, Y, YY, SY, v, gamma, start_index)
    m = length(S);
    for i = 1 : m
        ind = i - 1 + start_index;
        if(ind > m)
            ind = ind - m;
        end
        YGS{i} = fns.TVaddTV(x, S{ind}, fns.STV(x, -gamma, Y{ind}));
    end
    M = SY - gamma * YY;
    V = zeros(m, 1);
    for i = 1 : m
        V(i) = fns.inpro(x, YGS{i}, v);
    end
    V = linsolve(M, V);
    output = fns.STV(x, gamma, v);
    for i = 1 : m
        output = fns.TVaddTV(x, output, fns.STV(x, V(i), YGS{i}));
    end
end


function output = HV(fns, x, S, Y, YY, SY, V, gamma, start_index)
    m = length(S);
    for i = 1 : m
        ind = i - 1 + start_index;
        if(ind > m)
            ind = ind - m;
        end
        YGS{i} = fns.TVaddTV(x, S{ind}, fns.STV(x, -gamma, Y{ind}));
    end
    M = SY - gamma * YY;
    V1 = zeros(m, length(V));
    for j = 1 : length(V)
        for i = 1 : m
            V1(i,j) = fns.inpro(x, YGS{i}, V{j});
        end
    end
    V1 = linsolve(M, V1);
    for i = 1 : length(V)
        output{i} = fns.STV(x, gamma, V{i});
    end
    for j = 1 : length(V)
        for i = 1 : m
            output{j} = fns.TVaddTV(x, output{j}, fns.STV(x, V1(i,j), YGS{i}));
        end
    end
end
    
   

function output = Get_eta(q, x1, S, Y, RHO, start_index, gamma, m, fns)
    n = length(RHO);
    if(length(RHO) < m)
        for i = n : -1 : 1
            xi(i) = RHO(i) * fns.inpro(x1, S{i}, q);
            q = fns.TVaddTV(x1, q, fns.STV(x1, -xi(i), Y{i}));
        end
        r = fns.STV(x1, gamma, q);
        for i = 1 : n
            omega = RHO(i) * fns.inpro(x1, Y{i}, r);
            r = fns.TVaddTV(x1, r, fns.STV(x1, (xi(i) - omega), S{i}));
        end
        output = fns.STV(x1, -1, r);
        return;
    end

    for i = start_index + m - 1 : -1 : start_index
        ind = i;
        if(ind > m)
            ind = ind - m;
        end
        xi(ind) = RHO(ind) * fns.inpro(x1, S{ind}, q);
        q = fns.TVaddTV(x1, q, fns.STV(x1, -xi(ind), Y{ind}));
    end
    r = fns.STV(x1, gamma, q);
    for i = start_index : start_index + m - 1
        ind = i;
        if(ind > m)
            ind = ind - m;
        end
        omega = RHO(ind) * fns.inpro(x1, Y{ind}, r);
        r = fns.TVaddTV(x1, r, fns.STV(x1, (xi(ind) - omega), S{ind}));
    end
    output = fns.STV(x1, -1, r);
end


