  function varargout = REsubGA(fns,params)
% Solver of Riemannian \epsilon-sungradient method
%
% INPUT:
% fns : a struct that contains required function handles
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
%     params.H0 : initial approximation of inverse Hessian.
% not required:
%     params.error [1e-4]        : tolerance of stopping criterion.
%     params.delta [1e-5]        : stopping criterion of the descent direction algorithm;                           
%     params.epsilon [1]         : initial \epsilon
%     params.theta_epsilon [0.1] : shrink \epsilon constant
%     params.theta_delta [1]     : shrink delta constant
%     params.c [0.005]           : acceptance or rejection constant. It is between [0, 1)
%     params.gamma [0.5]         : backtrack factor in line search algorithm. It is between [0, 1]
%     params.max_ls [5000]       : the maximum number of iterations in the linesearch algorithm
%     params.max_t [5000]        : the maximum number of iterations
% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
% 
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates of x
%     varargout{2} : all iterates of y
%     varargout{3} : function values of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : time of function evaluations
%     varargout{7} : time to solve a quadratic programming problem
%     varargout{8} : accumulated computational time for iterations
%     
% 
    fprintf('REsubGA\n')
    if nargin < 2
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'delta', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'error', 1e-4, 'float', 0, 1);
    params = set_default_para(params, 'epsilon', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'gamma', 0.5, 'float', 0, 1);
    params = set_default_para(params, 'theta_epsilon', 0.1, 'float', 0, 1); 
    params = set_default_para(params, 'theta_delta', 1, 'float', 0, 1);
    params = set_default_para(params, 'c', 0.005, 'float', 0, 1);
    params = set_default_para(params, 'max_ls', 5000, 'int', 0, inf);
    params = set_default_para(params, 'max_t', 5000, 'int', 0, inf);

    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.Hv = fs.Hv;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
        fns.R = fs.R;
        fns.inpro = fs.inpro;
        fns.Tranv = fs.Tranv;
        fns.invTranv = fs.invTranv;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'Hv');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'Tranv');
    check_function_handle(fns, 'invTranv');
    times = 0;
    delta = params.delta;
    epsilon = params.epsilon;
    gamma = params.gamma;
    max_ls = params.max_ls;
    theta_epsilon = params.theta_epsilon;
    theta_delta = params.theta_delta;
    c = params.c;   
    x1 = params.x0;
    f1 = fns.f(x1); 
    X{1} = x1;
    F(1) = f1;
    Y = [];
    tt=1;
    fprintf('times : %d, f(x) : %e\n', times, f1.f);
    tic
    num = 0;
    m = 1; 
    l = 1;
    nftime = 0;
    quadtime = 0;
    options = optimset('Display','off','TolX', 1e-6, 'TolFun', 1e-6, 'Algorithm', 'active-set');
    while(1)
        x_s = x1;
        ns = 1;
        flag = 0;
        while(l <= params.max_t)           
            [norm_w_s, w_s, g, g_s, num1, numftime, quadtime1, ls] = Decent(fns, f1, x_s, delta, c, epsilon, l, times, F, X, options);
            quadtime = quadtime + quadtime1;
            nftime = nftime + numftime;
            tt = tt + num1;
            if(f1.f - params.minf) / (abs(params.minf) + 1) <= params.error
                flag = 1;
                times = times + 1;
                F(times + 1) = f1;
                X{times + 1} = x1;
                T(times + 1) = toc;
                break;
            end
            if(epsilon <= 1e-5)
                flag = 1;
                break;
            else
                inv_norm_w_s = 1 / norm_w_s; 
                g_s = fns.STV(x_s, inv_norm_w_s, w_s);
            end
            if(norm_w_s <= delta)
                if(m > 4)
                    theta_epsilon = 0.8;
                end
                epsilon = epsilon * theta_epsilon;
                delta = delta * theta_delta;            
                x1 = x_s;
                num = num + 1; 
                m = m+1;
                lnf=0;
                lnftime = 0;          
            else
                [sigma, x2, f2, beta, gamma, nd, f1, lnf, lnftime] = linesearchgs(fns, x_s, g_s, c, gamma, norm_w_s, 1, max_ls, f1, epsilon);
                x_s = x2;
                f1 = f2;
                times = times + 1;
                F(times + 1) = f1;
                X{times + 1} = x1;
                T(times + 1) = toc;
            end  
            l = l + 1;
            nftime = nftime + lnftime;
            tt = tt + lnf;
            if(ls == 0)
                break;
            end
        end
        if(flag == 1 || l > 5000 || ls == 0)
            break;
        end
        
    end
    timecost = toc;      
    varargout{1} = X;
    varargout{2} = Y;
    varargout{3} = F;
    varargout{4} = timecost;
    varargout{5} = tt;
    varargout{6} = nftime;
    varargout{7} = quadtime;
    varargout{8} = T;
end



function [t, xnew, f2, beta, gamma, nd, f1, lnf, lnftime, ls] = linesearchgs(fns, x, d, beta, gamma, nd, t, max_ls, f1, epsilon)
% This function apply the linesearch algorithm to find a point.
%
% INPUT:
% d       : search direction
% nd      : norm of search direction
% gamma   : backtrack factor
% x       : current iterate
% f1      : previous function values, the last one is the function value at the current iterate
% t       : the initial stepsize
% beta    : acceptance or rejection constant
% epsilon : the minimum stepsize
% max_ls  : the maximum number of iterations in the linesearch algorithm
% fns : a struct that contains required function handles
%     fns.f(x) : return objective function value at x
%     fns.gf(x) : return the subgradient of objection function at x
%     
% OUTPUT:
% t        : desired stepsize
% xnew     : next iterate x + t * d
% alpha    : value of the subgardient locality measure 
% f2       : function value at the next iterate x + t * d
% f1       : previous function values
% beta     : value of the subgardient locality measure 
% gamma    : coefficient in the Amijo condition
% nd       : norm of search direction
% lnftime  : time of function evaluations in the linesearch algorithm
% lnf      : number of function evaluations in the linesearch algorithm
%
    xnew = fns.R(x, d);
    times = 0;   
    lnftimestart = toc;
    f2 = fns.f(xnew);
    lnftime = toc - lnftimestart;
    lnf = 1;
    while( times < max_ls )
        if(f2.f - f1.f <= - beta * t * nd && t >= epsilon)
            break;
        else       
           t = gamma * t;
           if(t < epsilon)
               t = epsilon;
           end               
           td = fns.STV(x, t, d);
           xnew = fns.R(x, td);
           lnftimestart1 = toc;
           f2 = fns.f(xnew);
           lnftime = lnftime + toc - lnftimestart1;
           lnf = lnf + 1;
           times = times + 1;
        end
    end
     
end


function [norm_w_k, w_k1, g, g_k1, num1, numftime, quadtime, ls] = Decent(fns, f1, x, delta, c, epsilon, l, times, F, X, options)
% find a descent direction.
%
% INPUT:
% x       : current iterate
% f1      : previous function values, the last one is the function value at the current iterate
% c       : acceptance or rejection constant
% delta   : stopping criterion
% epsilon : approximation parameter
% options : set quadratic programming parameters
% fns : a struct that contains required function handles
%     fns.f(x) : return objective function value at x
%     fns.gf(x) : return the subgradient of objection function at x
%     
% OUTPUT:
% w_k1     : solution of quadratic programming problem
% norm_w_k : norm of w_k1
% g        : tangent vector on T_x M
% g_k1     : desired descent direction
% num1     : number of function evaluations in the linesearch algorithm
% numftime : time of function evaluations in the linesearch algorithm
% quadtime : time to solve a quadratic programming problem
%
%
    [m,p] = size(x.U);
    g = fns.gf(x, f1);
    ing = 1 / sqrt(fns.inpro(x, g, g));
    g = fns.STV(x, ing, g);
    g_k1 = g;
    epsilon_g = fns.STV(x, epsilon, g);
    xnew = fns.R(x, epsilon_g);
    fnew ={};
    [v, epsilon_g, x, xnew] = fns.invTranv(x, epsilon_g, xnew, fns.gf(xnew,fnew));
    W=[];
    W(:,1) = reshape(v.TV, [], 1);
    k = 1;
    num1=0;
    numftime = 0;
    quadtime = 0;
    ls = 1;
    while(1)      
        [m,N] = size(W);
        H = W' * W; 
        e = ones(1,N);    
        O = zeros(1,N);
        w0 = ones(N,1)/N;
        quadtimestart = toc;
        lambda = quadprog(H, O', [], [], e, 1, O, [], w0, options);
        quadtime = quadtime + toc - quadtimestart;
        w_k = -W * lambda;
        w_k1.TV = reshape(w_k, size(g.TV));
        norm_w_k = sqrt(fns.inpro(x, w_k1, w_k1));
        if(norm_w_k <= delta)
            break;
        else
            inv_norm_w_k = 1/ norm_w_k;
            g_k1 = fns.STV(x, inv_norm_w_k, w_k1);
        end 
        [h_epsilon, xnew1, h_epsilontime]= h(fns, f1, epsilon, x, g_k1, c, norm_w_k);
        numftime = numftime + h_epsilontime;
        num1 = num1 + 1;
        if(h_epsilon <= 0)   
            break;
        end    
        g_old = g;
        g = g_k1;
        [v, xnew, t_g, t, n, nftime, ls] = Increasing(fns, f1, x, g, 0, epsilon, c, norm_w_k);
        [v, t_g, x, xnew] = fns.invTranv(x, t_g, xnew, v);
        W(:,k + 1) = reshape(v.TV, [], 1);
        k = k + 1;
        numftime = numftime + nftime;
        num1 = num1+ n;   
        if(ls == 0)
            fprintf('warning: Increasing fails!\n');
            break;
        end
    end    
end



function [v, xnew, t_g, t, n, nftime, ls] = Increasing(fns, f1, x, g, a, b, c, nw)
% an h - increasing point algorithm
%
% INPUT:
% nw  : norm of w
% g   : tangent vector on T_x M
% x   : current iterate
% f1  : previous function values
% b   : the initial stepsize
% c   : the Armijo parameter
% a   : the minimum stepsize
% fns : a struct that contains required function handles
%     fns.f(x) : return objective function value at x
%     fns.gf(x) : return the subgradient of objection function at x
%     
% OUTPUT:
% v      : desired subgradient
% xnew   : next iterate x + t * g
% t_g    : t * g
% t      : stepsize
% n      : number of function evaluations in the linesearch algorithm
% nftime : time of function evaluations in the linesearch algorithm
% 
%
    t = b; 
    n = 0;
    nftime = 0;
    ls = 1;
    while(1)       
        t_g = fns.STV(x, t, g);
        xnew = fns.R(x, t_g);
        fnew = {};
        v = fns.gf(xnew, fnew);
        [dexp_g, t_g, x, xnew] = fns.Tranv(x, t_g, xnew, g);
        v_dexp_g1 = fns.inpro(xnew, v, dexp_g);
        if(v_dexp_g1 + c * nw >= 0)
             break;
        else
            t = (a + b) / 2;
            [h_t, x_t, h_ttime]= h(fns, f1, t, x, g, c, nw);
            [h_b, x_b, h_btime]= h(fns, f1, b, x, g, c, nw);
            nftime = nftime + h_ttime + h_btime;
            n = n + 2;
            if(h_b > h_t)
                a = t;
            else
                b = t;
            end      
        end  
        if(b - a <= eps)
            fprintf('warning: increasing reaches the max num of iters \n')
            ls = 0;
            break;           
        end
    end
end


function [value, xnew, ftime] = h(fns, f1, t, x, g, c, nw)
% Compute h(t) = f(exp_x(tg_k)) - f(x) + c * t * \|w_k\|
%
% INPUT:
% g   : g_k
% nw  : norm of w_k
% x   : current iterate
% f1  : previous function values of f(x)
% t   : the initial stepsize
% c   : the Armijo parameter
% fns : a struct that contains required function handles
%     fns.f(x) : return objective function value at x   
% OUTPUT:
% ftime : time of function evaluations
% xnew  : next iterate x + t * d
% value : value of h(t)
%
    t_g = fns.STV(x, t, g);
    xnew = fns.R(x, t_g);
    ftimestart = toc;
    f_R = fns.f(xnew);
    ftime = toc - ftimestart;
    value = f_R.f - f1.f + c * t * nw;
end






