function varargout = RPBM(fns, params)
%Solver of Riemannian proximal bundle method
% fns
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x1 M to T_{R_x1(d)} M. Here x2 = R_x1(d).
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%                                  and return some parameters which are useful in fns.Tranv_Params.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
%     params.H0 : initial approximation of inverse Hessian.
% not required:
%     params.tol [1e-5]     : tolerance of stopping criterion.
%     params.error [1e-4]   : another tolerance of stopping criterion
%     params.epsilon [0.05] : injectivity radius
%     params.alpha [1.2]    : a positive parameter to set a dynamic adjustment parameter
%     params.m_L [0.0001]   : the maximum number of iterations.
%     params.mu_l [1]       : proximal parameter
%     params.max_L [50]     : the maximum bundle size.
%     params.minf           : the known optimal objective function value or the minimum value found by the used algorithms
%     params.max_t [5000]   : the maximum number of iterations
%     params.max [inf]      : the maximum number of function evaluations
%

% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates of x
%     varargout{2} : all iterates of y
%     varargout{3} : function values of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : time of function evaluations
%     varargout{7} : time to solve a quadratic programming problem
%     varargout{8} : accumulated computational time for iterations



    fprintf('RPBM\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'error', 1e-4, 'float', 0, 1);
    params = set_default_para(params, 'tol', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'epsilon', 0.05, 'float', 0, 1);
    params = set_default_para(params, 'max_t', 5000, 'int', 1, inf);
    params = set_default_para(params, 'max', inf, 'int', 0, inf);
    params = set_default_para(params, 'mu_l', 1, 'float', 1, inf);
    params = set_default_para(params, 'alpha', 1.2, 'float', 0, inf);
    params = set_default_para(params, 'max_L', 50, 'int', 1, inf);
    params = set_default_para(params, 'm_L', 0.0001, 'float', 0, 1);

    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))  % isfield (s'fieldname')ṹSǷfieldnameָ
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.Hv = fs.Hv;  %
        fns.R = fs.R;
        fns.inpro = fs.inpro;
        fns.Tranv = fs.Tranv;
        fns.beta = fs.beta;
        fns.invTranv = fs.invTranv;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'Hv');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'Tranv');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');

    % Initialization
    nn=0;
    times = 0;
    num = 1;
    x1 = params.x0;
    fx1 = fns.f(x1);
    xi1 = fns.gf(x1, fx1);
    alpha0 = 0;
    Y=[];
    C = [];
    E = [];
    Xi=[];
    Xi(:,1) = reshape(xi1.TV, [], 1);     
    X{1} = x1;
    F(1) = fx1;   
    nftime = 0;
    quadtime = 0;
    fprintf('times : %d, f(x) : %e\n', times, fx1.f);
    options = optimset('Display','off','TolX', 1e-6, 'TolFun', 1e-6, 'Algorithm', 'active-set');
    tic  % 
    while(nn < params.max_t && times < params.max)     
        if((fx1.f - params.minf)/(abs(params.minf)+1) <= params.error)
            break;
        end
        if(num > 1)
            [M, N] = size(Xi);
            H = 1 / params.mu_l * ((Xi)' * (Xi)); 
            H = (H + H') / 2;
            e = ones(1, N);  
            lb = zeros(N, 1);
            w0 = ones(N, 1) / N;
            
            quadtimestart = toc;
            lambdav = quadprogsub(H, C', [], [], e, 1, lb, e, w0,options);  
            quadtime = quadtime + toc - quadtimestart;
            c_lambdav = C * lambdav;
            
            [n, p] = size(lambdav);
            xi_lambda.TV = reshape(Xi * lambdav, M, 1);
            norm_xi_lambda = fns.inpro(x1, xi_lambda, xi_lambda);
            vk_1 = 1 / params.mu_l * norm_xi_lambda + c_lambdav;
            inv_mu_1 = -1 / params.mu_l;
            dk_1 = fns.STV(x1, inv_mu_1, xi_lambda);
        else
            eta_1 = max(alpha0, params.alpha) + params.alpha;
            norm_xi1 = fns.inpro(x1, xi1, xi1);
            vk_1 = 1 / params.mu_l * norm_xi1;
            inv_mu_1 = -1 / params.mu_l;
            dk_1 = fns.STV(x1, inv_mu_1, xi1);
        end
            
        if(vk_1 <= params.tol)
            break;
        end
        norm_dk_1 = sqrt(fns.inpro(x1, dk_1, dk_1));
        if(norm_dk_1 <= params.epsilon)         
            g_1 = dk_1;
            y2 = fns.R(x1, g_1);
            alpha0 = 0;
            nftimestart = toc;
            fy2 = fns.f(y2);
            nftime = nftime + toc - nftimestart;
            xi1 = fns.gf(y2,fy2);
            [Txi1, g_1, x1, y2] = fns.invTranv(x1, g_1, y2, xi1);
        else
            inv_norm_dk_1 = 1 / norm_dk_1;         
            g_1 = fns.STV(x1, inv_norm_dk_1, dk_1); 
            g_1 = fns.STV(x1, params.epsilon, g_1);
            y2 = fns.R(x1, g_1);
            nftimestart = toc;
            fy2 = fns.f(y2);
            nftime = nftime + toc - nftimestart;
            xi1 = fns.gf(y2,fy2);      
            [Txi1, g_1, x1, y2] = fns.invTranv(x1, g_1, y2, xi1);
            alpha0 = -fns.inpro(x1, Txi1, g_1) * inv_norm_dk_1 / params.epsilon ;
                   
        end           
        
        if(fy2.f < fx1.f - params.m_L * vk_1)   %  serious step
            x1 = y2;
            fx1 = fy2;
            Xi=[];
            Xi(:,1) = reshape(xi1.TV, [], 1);
            C = []; 
            E= [];    
            num = 1;
            nn = nn + 1;
            X{nn + 1} = x1;
            F(nn + 1) = fx1;
            T(nn + 1) = toc;        
        else     %  null step
            norm_g_1 = fns.inpro(x1, g_1, g_1);
            num = num + 1;
            ek = fx1.f - fy2.f + fns.inpro(x1, Txi1, g_1);
            if(num <= params.max_L)
                E(:,num - 1) = ek / norm_g_1;
            else
                E(:, 1 : params.max_L - 2) = E(:,2 : params.max_L-1);
                E(:, params.max_L - 1) = ek / norm_g_1;
            end
            eta_1 = max([max(-2 * E), alpha0, params.alpha]) + params.alpha;
            ck = ek + eta_1 / 2 * norm_g_1;
            xik = fns.TVaddTV(x1, Txi1, fns.STV(x1, eta_1, g_1));           
            if(num <= params.max_L)
                Xi(:,num) = reshape(xik.TV, [], 1);
                C(:,num) = ck;
            else
                Xi(:, 1 : params.max_L-1) = Xi(:, 2 : params.max_L);
                C(:, 1 : params.max_L-1) = C(:, 2 : params.max_L);
                Xi(:, params.max_L) = reshape(xik.TV, [], 1);
                C(:, params.max_L) = ck;
            end           
        end  
        times = times + 1;
    end
    timecost = toc;
    varargout{1} = X;
    varargout{2} = Y;  
    varargout{3} = F;
    varargout{4} = timecost;
    varargout{5} = times;
    varargout{6} = nftime;
    varargout{7} = quadtime;
    varargout{8} = T;

end   

    
    
