function varargout = RQNBM(fns, params)
%Solver of Riemannian quasi-newton bundle method
% fns
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x1 M to T_{R_x1(d)} M. Here x2 = R_x1(d).
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%                                  and return some parameters which are useful in fns.Tranv_Params.
%     fns.TranH(x1, d, x2, H1) : return a operator, H2, that affects on T_x2 M. H2 satisfies that for any v \in T_x2(M), H2(v) = Tran (H1 (Tran^{-1}(v)))
%                             where Tran is the vector transport fns.Tranv(x1, d, x2, v).
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
%     fns.operadd(x, H1, H2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = H1(v) + H2(v).
%     fns.rank1operator(x, v1, v2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = g_x(v2, v) * v1.
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
%     params.H0 : initial approximation of inverse Hessian.
% not required:
%     params.err [1e-5]         : tolerance of stopping criterion.
%     params.error [1e-4]       : another tolerance of stopping criterion
%     params.m [8]              : number of s and u for storage
%     params.max_t [5000]       : the maximum number of iterations.
%     params.update [1]         : '0' means that updates to Hk are skipped, '1' means update Hk.
%     params.rho [0.1 10-3]     : correction parameter
%     params.minf               : the known optimal objective function value or the minimum value found by the used algorithms
%     params.E [1]              : control direction length
%     params.mu0 [0.18]         : injectivity radius
%     params.Gamma [50]         : set the parameter to reduce SR1 updates
%     params.maxstepsize [1]    : allowed maximum stepsize 
%

% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates of x
%     varargout{2} : all iterates of y
%     varargout{3} : function values of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : time of function evaluations
%     varargout{7} : time to solve a quadratic programming problem
%     varargout{8} : accumulated computational time for iterations
%     varargout{9} : part of the values of wk


    fprintf('RQNBM\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'm', 8, 'int', 0, inf);
    params = set_default_para(params, 'error', 1e-4, 'float', 0, 1);
    params = set_default_para(params, 'err', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'rho', 0.1, 'float', 0, 1);
    params = set_default_para(params, 'max_t', 5000, 'int', 1, inf);
    params = set_default_para(params, 'update', 1, 'int', 0, 1);
    params = set_default_para(params, 'maxstepsize', 1, 'float', 1, inf);
    params = set_default_para(params, 'mu0', 0.18, 'float', 0, inf);
    params = set_default_para(params, 'E', 1, 'float', 0, inf);
    params = set_default_para(params, 'Gamma', 50, 'int', 1, inf);


 fprintf('RQNBM\n')
    if nargin < 2
       error('Invalid arguments: the number of arguments is invalid.');
    end
    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))  % isfield (s'fieldname')ṹSǷfieldnameָ
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.Hv = fs.Hv; 
        fns.R = fs.R;
        fns.inpro = fs.inpro;
        fns.Tranv = fs.Tranv;
        fns.beta = fs.beta;
        fns.TranH = fs.TranH;
        fns.invTranv = fs.invTranv;
        fns.rank1operator = fs.rank1operator;
        fns.operadd = fs.operadd;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'Hv');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'Tranv');
    check_function_handle(fns, 'invTranv');
    check_function_handle(fns, 'TranH');
    check_function_handle(fns, 'beta');
    check_function_handle(fns, 'rank1operator');
    check_function_handle(fns, 'operadd');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');

    % Check whether required parameters are legal
    if(~isfield(params, 'x0') || ~isfield(params, 'H0'))
        error('Invalid arguments: missing initial x0 or H0');
    end
    % Initialization
    times = 0;
    tt =1;
    x1 = params.x0;
    H = params.H0;
    [m, n] = size(H);
    D = params.D;
    maxstepsize = params.maxstepsize;
    f1 = fns.f(x1);
    num = 0;
    gradfx1 = fns.gf(x1,f1);

    cgradfx1 = gradfx1;
    w1 = fns.inpro(x1, cgradfx1, cgradfx1);  
%     W(1) = w1;
    y1 = x1;
    c_alpha = 0;
    
    X{1} = x1;
    Y{1} = y1;
    F(1) = f1;
    G(:,1) = reshape(gradfx1.TV,[],1);  
    iU = 0;
    iC = 0;
    nC = 0;
 
    nftime = 0;
    quadtime = 0;
%     N=[];
    fprintf('times : %d, f(x) : %e\n', times, f1.f);
    T(1)=0;
%     Q(1)=0;
%     NF(1)=0;
%     OT(1)=0;
    qtime = 0;
    options1 = optimset('Display','off','TolX', 1e-5, 'TolFun', 1e-5, 'Algorithm', 'active-set');
    options2 = optimset('Display','off','TolX', 1e-6, 'TolFun', 1e-6, 'Algorithm', 'active-set');
    tic 
    while(times< params.max_t)
        if((f1.f - params.minf)/(abs(params.minf)+1) <= params.error)
            break;
        end
        if(w1<params.err)
           break;
        end
        % direction      
        eta = fns.Hv(x1, H, cgradfx1);
        eta = fns.STV(x1, -1, eta);
        
        if(times >= params.t)
            params.rho = 1e-3;
        end                                            
        n_eta = sqrt(fns.inpro(x1, eta, eta));% norm  
        initstepsize = min([maxstepsize, params.mu0 / n_eta]);        
        % line search
        [new_eta, Tgfnew, t_R, t_L, alpha, x2, y2, fx2, fy2, gfnew, ins, beta, lnf, lnftime] = linesearch_R(fns, eta, x1, f1, w1, n_eta, initstepsize, params, D);

        if(t_L>0) %serious step
            gradfx1 = gfnew;
            G=[];
            G(:,1) = reshape(gradfx1.TV,[],1);
            cgradfx1 = gfnew;
            c_alpha = 0;
            n_cgradfx1 = fns.inpro(x2, cgradfx1, cgradfx1); 
            
            tilde_H = fns.TranH(x1, new_eta, x2, H);  
            if(params.update ==1)
                [Tg, new_eta, x1, y2] = fns.Tranv(x1, new_eta, y2, gradfx1); 
                u_k = fns.TVaddTV(y2, gfnew, fns.STV(y2, -1, Tg));  
                s_k = fns.Tranv(x1, new_eta, x2, new_eta);    
                inp_u_s = fns.inpro(x2, u_k, s_k);    
                % BFGS update
                if(inp_u_s > params.rho)  
                    Hu_k = fns.Hv(x2, tilde_H, u_k);  
                    check_H = fns.operadd(x2, tilde_H, fns.rank1operator(x2, fns.STV(x2, -1 / inp_u_s, s_k), Hu_k));
                    Hu_k = fns.Hv(x2, check_H, u_k);
                    check_H = fns.operadd(x2, check_H, fns.rank1operator(x2, Hu_k, fns.STV(x2, - 1 / inp_u_s, s_k)));          
                    check_H = fns.operadd(x2, check_H, fns.rank1operator(x2, fns.STV(x2, 1 / inp_u_s, s_k), s_k));          
                    iU = 1;
                else
                    check_H = tilde_H;
                    iU = 0;

                end
            else
                check_H = tilde_H;
            end
        else
            if(params.update ==1)
                u_k = fns.TVaddTV(x2, Tgfnew, fns.STV(x1, -1, gradfx1));
                v_k = fns.TVaddTV(x2, fns.Hv(x1, H, u_k), fns.STV(y2, -1, new_eta));
                inp_u_v = fns.inpro(x2, u_k, v_k);
                inp_cg_v = fns.inpro(x2, cgradfx1, v_k);
            end
            G(:,2) = reshape(Tgfnew.TV,[],1);
            G(:,2) = fns.beta(x1, gfnew, x2) * G(:,2);

            re_cgradfx1 = reshape(cgradfx1.TV,[],1);
            % quadratic programming
            if(G(:,1)~=re_cgradfx1)
                G(:,3) = reshape(cgradfx1.TV,[],1);
                H1= G'* H * G;
                H1 =(H1 + H1')/2;
                c = [0; alpha; c_alpha]; 
                e = ones(1,3);  
                w0 = ones(3,1)/3;
                O = zeros(3,1);                
                quadtimestart = toc;
                lambdav = quadprog(H1, c, [], [], e, 1, O, [], w0, options1); 
                qtime =  qtime+ toc - quadtimestart;
                quadtime =quadtime + toc - quadtimestart ;
                c_alpha = lambdav' * c;
                lambda.TV = reshape(lambdav,[],1);           
                cgradfx1 = fns.Hv(x2, G, lambda);
                G(:,3) = [];
            else                 
                H1= G'* H * G;
                H1 =(H1 + H1')/2;
                c = [0; alpha]; 
                e = ones(1,2); 
                w0 = ones(2,1)/2;
                O = zeros(2,1);                
                quadtimestart = toc;
                lambdav = quadprog(H1, c, [], [], e, 1, O, [], w0,options2);  %
                qtime = qtime+ toc - quadtimestart;
                quadtime = quadtime + toc - quadtimestart;
                c_alpha = lambdav' * c;
                lambda.TV = reshape(lambdav,[],1);           
                cgradfx1 = fns.Hv(x2, G, lambda);
            end            
            % SR1 update           
            n_cgradfx1 = fns.inpro(x2, cgradfx1, cgradfx1);
            if(params.update ==1)
                if( inp_cg_v <0 && (iC==0 || (params.rho * n_cgradfx1 <= inp_cg_v^2 / inp_u_v && params.rho * n <= fns.inpro(x2, v_k, v_k) / inp_u_v)))
                    check_H = fns.operadd(x2, H, fns.rank1operator(x2, fns.STV(x2, -1 / inp_u_v, v_k), v_k));  
                    iU = 1;
                else
                    check_H = H;
                    iU = 0;
                end  
            else
                check_H = H;
            end
        end
        
        % Correction
        cHcg = fns.Hv(x2, check_H, cgradfx1);
        n_H_eta = sqrt(fns.inpro(x2, cHcg, cHcg));   %norm
        if(n_H_eta > params.E)    %control direction length           
            check_H = (params.E / n_H_eta) * check_H;
            cHcg = fns.Hv(x2, check_H, cgradfx1);
%             cHcg = fns.STV(x2, params.E / n_H_eta, cHcg);
        end
        % compute w1 
        check_w1 = fns.inpro(x2, cgradfx1, cHcg) + 2 * c_alpha;        
        if(check_w1 < params.rho * n_cgradfx1 || (iC ==1 && iU == 1))
            w1 = check_w1 + params.rho * n_cgradfx1;
            H = fns.operadd(x2, check_H, params.rho * eye(n)); 
            nC = nC +1;              
        else
            w1 = check_w1;    
            H = check_H;
        end

        if(nC >= params.Gamma)  
            iC = 1;
        end
        
        tt = tt + lnf ;
        nftime = nftime + lnftime;
        times = times + 1;  
%         W1(times + 1) = w1;      
        if(x1.U ~= x2.U)    
            num =num +1;
            x1 = x2;
            y1 = y2;
            f1 = fx2;      
            X{num + 1} = x1;
            Y{num + 1} = y1;
            F(num + 1) = f1;
            T(num + 1) = toc;
%             W(num + 1) = W1(times);
%             Q(num + 1) = quadtime;
%             NF(num + 1) = nftime;
%             qtime = 0;
%             OT(num + 1) = toc;
        end 
    end
    timecost = toc;
    varargout{1} = X;
    varargout{2} = Y;  
    varargout{3} = F; 
    varargout{4} = timecost;
    varargout{5} = tt;
    varargout{6} = nftime;
    varargout{7} = quadtime;
    varargout{8} = T;
%     varargout{9} = W;
%     varargout{10} = Q;
%     varargout{11} = NF;
%     varargout{12} = OT;
end   
    
    
