 function [new_eta, Tgf, step_size_R, step_size_L, alpha, xnew1, xnew, fnew2, fnew, gfnew, ins, delta, lnf ,lnftime] = linesearch_R(fns, eta, x, fv, w1, neta, initialstepsize, params, D)
 % This function apply the linesearch algorithm to find a point.
%
% INPUT:
% eta : search direction
% neta : norm of search direction
% w1  : wk (stopping criterion) at the current iterate 
% x   : current iterate
% fv  :  previous function values, the last one is the function value at the current iterate
% gradf1 : gradient at the current iterate
% initialstepsize : the initial step size
% D  : control the distance between y and x in null step
% fns : a struct that contains required function handles
%     fns.f(x) : return objective function value at x
%     fns.gf(x) : return the subgradient of objection function at x
%     
% params: : a struct that contains parameters that are used
%     theta_A : coefficient in the serious step
%     kappa : the interval parameter (0,1/2).
%     theta_R : coefficient in the null step
%     theta_T : determine the next iteration stepsize interval
%     theta_L : coefficient in the serious step
%     gamma : correspond to the measure of nonconvexity of f
%     omega : locality measure parameter
%     minstepsize: allowed minimum stepsize
%
% OUTPUT:
% new_eta : step size * search direction
% Tgf : a tangent vector on T_x M, which is given by inverse vector transport that transport gf \in T_xnew M 
% stepsize : desired step size
% xnew  : next iterate y + new_eta
% step_size_R : desired stepsize for {x}
% step_size_L : desired stepsize for {y}
% xnew1 : next iterate x + new_eta
% alpha : value of the subgardient locality measure 
% fnew2: function value at the next iterate y + new_eta
% gfnew : subgradient at the next iterate y + new_eta
% fnew : function value at the next iterate x + new_eta
% ins :  the inner product of Tgf and eta
% delta : value of the subgardient locality measure    
% lnftime : time of function evaluations in the linesearch algorithm
% lnf : number of function evaluations in the linesearch algorithm
    theta_A = params.thetaA;
    theta_L = params.thetaL;
    theta_R = params.thetaR;
    theta_T = params.thetaT;
    minstepsize = params.minstepsize;
   

    step_size_a = 0;
    step_size_u = initialstepsize;
    step_size = initialstepsize;
    lnf = 0;
    lnftime = 0;

    while(1)         
        new_eta = fns.STV(x, step_size, eta);
        xnew = fns.R(x, new_eta);
        lnftimestart = toc;
        fnew = fns.f(xnew);
        lnftime = lnftime + toc - lnftimestart;
        lnf = lnf + 1;
        gfnew = fns.gf(xnew, fnew);
        
        [Tgf, new_eta, x, xnew] = fns.invTranv(x, new_eta, xnew, gfnew);
        ins = 1/fns.beta(x, gfnew, xnew) * fns.inpro(x, Tgf, eta);

        delta = max([abs(fv.f - fnew.f + step_size * ins), params.gamma * (step_size * neta)^params.omega]);
        if(fnew.f <= fv.f - theta_T * step_size * w1)
            step_size_a = step_size;
            
        else
            step_size_u = step_size;         
        end 
        
        if(fnew.f <= fv.f - theta_L * step_size * w1 && (step_size >= minstepsize || delta > theta_A * w1)) % serious 
            step_size_R = step_size;
            step_size_L = step_size;
            alpha = 0;
            xnew1 = xnew;
            fnew2 = fnew;
            return;        
        else
            if(ins - delta >= -theta_R * w1 && (step_size - step_size_a) * neta < D) %null
                step_size_R = step_size;
                step_size_L = 0;
                alpha = delta;
                xnew1 = x;
                fnew2 = fv;
                return; 
            end
        end

        left_t = step_size_a + params.kappa * (step_size_u - step_size_a);
        right_t = step_size_u - params.kappa * (step_size_u - step_size_a);
        step_size = (left_t + right_t)/4;      
    end
end

